#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#ifdef ARCHI_NetBSD
# include <sys/dirent.h>
#else /* !ARCHI_NetBSD */
# include <dirent.h>
#endif /* !ARCHI_NetBSD */
#include "browse.h"
#include "list/s_list.h"
#include "fileinfo.h"

struct			s_spth
{
  const char		*sfind;
  const char		*path;
};

int			perror_fail(const char		*sfind,
				    const char		*path)
{
  write(STDERR_FILENO, sfind, strlen(sfind));
  write(STDERR_FILENO, ": ", 2);
  perror(path);
  return 1;
}

static int		check_match(const char		*fn,
				    const char		*absfn)
{
  absfn = absfn;
  if (!strncmp(fn, "Makefile", 8))
    return 0;
  if (!strcmp(fn, "directives.conf"))
    return 0;
  if (fn[0] == '.')
    return 0;
  if (fn[strlen(fn)-1] == '~')
    return 0;
  return 1;
}

static struct s_list*	browse_rec(struct s_list *res,
				   const char	*sfind,
				   int		len_opath,
				   const char	*path);

static struct s_list*	browse_file(struct s_list*res,
				    struct s_spth	*spth,
				    int		len_opath,
				    DIR		*dir)
{
  struct dirent		*dp;
  struct stat		sb;
  struct fileinfo	*fi;
  char			*ext;

  while ((dp = readdir(dir)))
    {
      fi = malloc(sizeof(struct fileinfo));
      fi->dn = strdup(spth->path);
      fi->bn = strdup(dp->d_name);
      fi->absfn = strdup(fi->dn);
      fi->absfn =
	realloc(fi->absfn,
		(strlen(fi->absfn) + 1 + strlen(fi->bn) + 1) *
		sizeof(char));
      fi->absfn = strcat(fi->absfn, "/");
      fi->absfn = strcat(fi->absfn, fi->bn);
      for (ext = fi->bn + strlen(fi->bn);
	   ext != fi->bn && *ext != '.' && *ext != '/';
	   ext--)
	;
      if (*ext == '/')
	fi->ext = strdup("");
      else
	if (*ext == '.')
	  fi->ext = strdup(ext);
	else
	  fi->ext = strdup("");
      fi->reldn = strdup(fi->dn + len_opath);
      fi->relfn = strndup(fi->absfn + len_opath + 1, strlen(fi->absfn) - len_opath - 1);
      fi->relfn_noext = strndup(fi->absfn + len_opath + 1, strlen(fi->absfn) - len_opath - strlen(fi->ext) - 1);
      fi->bn_noext = strndup(fi->bn, strlen(fi->bn) - strlen(fi->ext));
      if (lstat(fi->absfn, &sb) == -1)
	{
	  free_fileinfo(fi);
	  return NULL;
	}
      if (!(strcmp(fi->bn, ".") && strcmp(fi->bn, "..")))
	free_fileinfo(fi);
      else
	if (!check_match(fi->bn, fi->absfn))
	  free_fileinfo(fi);
	else
	  if (S_ISDIR(sb.st_mode) && !S_ISBLK(sb.st_mode) &&
	      !S_ISLNK(sb.st_mode) && !S_ISSOCK(sb.st_mode))
	    {
	      res = browse_rec(res, spth->sfind, len_opath, fi->absfn);
	      free_fileinfo(fi);
	    }
	  else
	    list_add_item_head(&res, fi);
    }
  return res;
}

static struct s_list*	browse_rec(struct s_list* res,
				   const char	*sfind,
				   int		len_opath,
				   const char	*path)
{
  struct s_spth		spth;
  DIR			*dir;

  if (!(dir = opendir(path)))
    {
      perror_fail(sfind, path);
      return NULL;
    }
  spth.path = path;
  spth.sfind = sfind;
  res = browse_file(res, &spth, len_opath, dir);
  if (closedir(dir) == -1)
    return NULL;
  return res;
}

struct s_list*	browse(const char		*sfind,
		       const char		*path)
{
  struct stat	sb;
  char		*fn;
  struct s_list* res = NULL;

  if (lstat(path, &sb) == -1)
    {
      return NULL;
    }
  fn = strdup(path);
  fn = realloc(fn, (strlen(fn) + 2) * sizeof(char));
  fn = strcat(fn, "/");
  if ((S_ISDIR(sb.st_mode) && !S_ISBLK(sb.st_mode) &&
       !S_ISLNK(sb.st_mode) && !S_ISSOCK(sb.st_mode)))
    {
      free(fn);
      return browse_rec(res, sfind, strlen(path), path);
    }
  free(fn);
  return res;
}
